package contrasts;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This is the main class of the contrast package.
 * 
 * The class has several tasks. Configurations can be set up here (e.g. the
 * alpha value). It stores values (the multinomial population), the resulting
 * confidence intervals, population sizes etc. And it provides several methods
 * to get informations, like the confidence intervals of a specific class,
 * scores of several population splits and so on.
 * 
 * @author <a href="mailto:max.schubach@informatik.uni-tuebingen.de">Max
 *         Schubach</a>
 * 
 * @version 1.1
 * 
 */
public class Contrasts {

	/**
	 * The methods used for this contrast
	 */
	private Method method;

	/**
	 * The multinomial population for the statistical test. Map must be
	 * Map<population, Map<class, [assigned values, summarized values]>>
	 */
	private Map<String, Map<String, Double>> values;

	/**
	 * The resulting confidence intervals (after run apply()). Map is stored in
	 * this way: Map<population1, Map<population2, Map<class, [assigned values
	 * interval, summarized values interval, assigned frequency, summarized
	 * frequency]>>>
	 * 
	 * Contrast generated in this way: population1(class) - population2(class)
	 */
	private Map<String, Map<String, Map<String, ConfidenceInterval>>> confidenceIntervals;

	/**
	 * List of contrasts include zero in their assigned value confidence
	 * interval. It is stored by a comparison by two populations.
	 */
	private Map<String, Map<String, ArrayList<String>>> contrastsIncludeZero;

	/**
	 * List of contasts include zero in their summarized value confidence
	 * interval. It is stored by a comparison by two populations.
	 */
	private Map<String, Map<String, ArrayList<String>>> contrastsExcludeZero;

	/**
	 * List of classes include zero at least one time.
	 */
	private ArrayList<String> classesIncludeZero;

	/**
	 * List of classes include zero at least one time.
	 */
	private ArrayList<String> classesExcludeZero;

	/**
	 * Population sizes. Computed after running apply() on a multinomial
	 * proportion. Needed for computation of intervals.
	 */
	private Map<String, Double> populationSizes;

	/**
	 * Floag for the small variance filter. If true, small variances are
	 * deleted.
	 */
	private boolean smallVarianceFilter;

	/**
	 * The total number of different populations.
	 */
	private int numberOfPopulations;

	/**
	 * The total number of different classes;
	 */
	private int numberOfClasses;

	/**
	 * Default constructor of contrasts. Sets the alpha value to 0.05 (default)
	 */
	public Contrasts() {
		this.method = new Method();
	}

	/**
	 * Setter method of alpha.
	 * 
	 * @param alpha
	 *            the wanted alpha
	 */
	public void setAlpha(double alpha) {
		this.method.setAlpha(alpha);
	}

	/**
	 * Setter method of the Confidence Interval method
	 * 
	 * @param method
	 *            The method of constructing intervals. Possible methods:
	 *            <ul>
	 *            <li>GOODMAN</li>
	 *            <li>BONFERRONI</li>
	 *            <li>GOODMAN+BONFERRONI</li>
	 *            <li>HOLM-BONFERRONI</li>
	 *            </ul>
	 */
	public void setMethod(String method) {
		this.method.setMethod(method.toUpperCase());

	}

	/**
	 * Gives all possible methods that are implemented for generating contrasts.
	 * 
	 * @return All possible implemented methods
	 */
	public String[] getMethods() {
		return (String[]) this.method.methods.toArray();

	}

	/**
	 * Setter method to use the Variance filter.
	 * 
	 * @param smallVarianceFilter
	 */
	public void setSmallVarianceFilter(boolean smallVarianceFilter) {
		this.smallVarianceFilter = smallVarianceFilter;
	}

	/**
	 * Getter method of confidence intervals.
	 * 
	 * @see ConfidenceInterval
	 * 
	 * @return all confidence intervals of all values Output is
	 *         {@code Map<population1, Map<population2, Map<class, assigned
	 *         interval>>}. Contrast is defined by
	 *         population1(class)-population2(class)
	 */
	public Map<String, Map<String, Map<String, ConfidenceInterval>>> getConfidenceIntervals() {
		return this.confidenceIntervals;
	}

	/**
	 * Getter method of classes that include zero in the assigned values
	 * confidence interval.
	 * 
	 * @return ArrayList of classes include zero. It is stored by a comparison
	 *         by two populations (map).
	 */
	public Map<String, Map<String, ArrayList<String>>> getClassesIncludeZero() {
		return this.contrastsIncludeZero;
	}

	/**
	 * Getter method of classes that do not include zero in the assigned values
	 * confidence interval.
	 * 
	 * @return ArrayList of classes where the confidence interval is different
	 *         from zero. It is stored by a comparison by two populations (map).
	 */
	public Map<String, Map<String, ArrayList<String>>> getClassesExcludeZero() {
		return this.contrastsExcludeZero;
	}

	/**
	 * Apply a multinomial population and create the confidence intervals.
	 * 
	 * First this method creates the populations sizes, second the frequencies,
	 * and third the confidence intervals are computed.
	 * 
	 * 
	 * @param values
	 *            An multinomial population. The map should be stored in this
	 *            way:
	 *            {@code Map<Population, Map<Class, [assigned values of class,
	 *            summarized values of class]>>}
	 * @param populationSizes
	 *            For every population the complete size. This is necessary if
	 *            there are internal summarized nodes.
	 */
	public void apply(Map<String, Map<String, Double>> values,
			Map<String, Double> populationSizes) {
		this.populationSizes = populationSizes;
		this.values = values;
		setNumberOfPopulations();
		setNumberOfClasses();
		this.method
				.setConstants(this.numberOfPopulations, this.numberOfClasses);
		System.err.println("Contrasts - Compute frequencies");
		if (this.smallVarianceFilter) {
			System.err.println("Contrasts - Apply small variance filter");
			applySmallVarianceFilter();
			System.err
					.println("Contrasts - Recalculating population sizes & frequencies");
			computePopulationSizes();
			setNumberOfPopulations();
			setNumberOfClasses();
			this.method.setConstants(this.numberOfPopulations,
					this.numberOfClasses);
		}
		computeFrequencies();
		System.err.println("Contrasts - Compute Confidence Intervals");
		computeConfidenceIntervals();
		System.err.println("Contrasts - Computation of intervals ready");
	}

	/**
	 * 
	 */
	private void setNumberOfClasses() {
		this.numberOfClasses = this.values.entrySet().iterator().next()
				.getValue().size();
	}

	/**
	 * 
	 */
	private void setNumberOfPopulations() {
		this.numberOfPopulations = this.values.size();
	}

	/**
	 * This Method filters the data sets by small variances. Therefore the
	 * average mean of the contrast variance is computed. A new mean is
	 * generated of these values and all classes that are below that mean are
	 * removed.
	 * 
	 * WARNING: After that the populations sizes have to be recomputed! That
	 * only works if no summarized inner node is available. Otherwise the new
	 * computed population size will be to large and the new contrasts test will
	 * be false.
	 * 
	 */
	private void applySmallVarianceFilter() {

		int combinations = this.populationSizes.size()
				* (this.populationSizes.size() - 1);

		Map<String, Double> variances = new HashMap<String, Double>();

		for (Map.Entry<String, Map<String, Double>> population_1_entry : this.values
				.entrySet()) {
			String populationName1 = population_1_entry.getKey();
			for (Map.Entry<String, Map<String, Double>> population_2_entry : this.values
					.entrySet()) {
				String populationName2 = population_2_entry.getKey();
				if (populationName1.equals(populationName2)) {
					continue;
				}
				// compute contrast & delta
				for (Map.Entry<String, Double> class_entry : population_1_entry
						.getValue().entrySet()) {
					String class_name = class_entry.getKey();
					double p1 = class_entry.getValue()
							/ this.populationSizes.get(populationName1);
					double p2 = 0.0;
					if (population_2_entry.getValue().containsKey(class_name)) {
						p2 = population_2_entry.getValue().get(class_name)
								/ this.populationSizes.get(populationName2);
						;
					}

					double variance = (p1 - (p1 * p1))
							/ this.populationSizes.get(populationName1)
							+ (p2 - (p2 * p2))
							/ this.populationSizes.get(populationName2);
					if (variances.containsKey(class_name)) {
						double new_variance = variances.get(class_name)
								+ variance;
						variances.put(class_name, new_variance);
					} else {
						variances.put(class_name, variance);
					}

				}
			}
		}

		double mean = 0.0;
		for (String class_name : variances.keySet()) {
			double mean_variance = variances.get(class_name) / combinations;
			mean += mean_variance;
			variances.put(class_name, mean_variance);
		}
		mean = mean / variances.size();

		// ValueComparator bvc = new ValueComparator(variances);
		// TreeMap<String, Double> sortedVariances = new TreeMap<String,
		// Double>(bvc);
		//
		// sortedVariances.putAll(variances);
		// int count = 0;
		// int removed = 0;
		// int i = -1;
		// for (String class_name : sortedVariances.keySet()) {
		// i++;
		// if (i < sortedVariances.size() * 0.10 ) {
		// System.out.println(sortedVariances.get(class_name));
		// count++;
		// continue;
		//
		// }
		// removeClass(class_name);
		// removed++;
		// }
		// System.out.println("Removing " + removed + " classes of total "
		// +(count+removed));
		//

		// PrintOut.sortedSplits(variances);
		int count = 0;
		int removed = 0;
		for (Map.Entry<String, Double> variance_entry : variances.entrySet()) {
			if (variance_entry.getValue() < mean) {
				removeClass(variance_entry.getKey());
				removed++;
			} else {
				count++;
			}
		}

		System.out.println("Removing " + removed + " classes of total "
				+ (count + removed));

	}

	/**
	 * This method removes a class completely from the value set in every
	 * population.
	 * 
	 * @param className
	 *            which will be removed.
	 */

	private void removeClass(String className) {
		for (String dataset : this.values.keySet()) {
			if (this.values.get(dataset).containsKey(className)) {
				this.values.get(dataset).remove(className);
			}
		}
	}

	/**
	 * Apply a multinomial population and create the confidence intervals.
	 * 
	 * First this method creates the populations sizes, second the frequencies,
	 * and third the confidence intervals are computed.
	 * 
	 * 
	 * @param values
	 *            An multinomial population. The map should be stored in this
	 *            way:
	 *            {@code Map<Population, Map<Class, [assigned values of class,
	 *            summarized values of class]>>}
	 */
	public void apply(Map<String, Map<String, Double>> values) {
		this.values = values;
		System.err.println("Contrasts - Compute population sizes");
		computePopulationSizes();
		this.apply(values, this.populationSizes);
	}

	/**
	 * Get the confidence intervals of specific contrast between a class of two
	 * populations.
	 * 
	 * @param dataset1_name
	 *            Name of the first dataset
	 * @param dataset2_name
	 *            Name of the second dataset
	 * @param class_name
	 *            Name of the class
	 * @return Returns a confidence interval generated by the contrast
	 *         population1(class) - population2(class). On position 0 the
	 *         assigned value contrast is stored. On position 1 the summarized
	 *         one.
	 */
	public ConfidenceInterval getClassInterval(String dataset1_name,
			String dataset2_name, String class_name) {

		return this.confidenceIntervals.get(dataset1_name).get(dataset2_name)
				.get(class_name);

	}

	/**
	 * Get the confidence intervals of all contrasts between a class for every
	 * population.
	 * 
	 * @param class_name
	 *            Name of the class
	 * @return All confidence intervals (CIs) of the given class. Provided by a
	 *         map: {@code Map<population1, Map<population2, CI}. Contrast is
	 *         population1(class) - population2(class)
	 */
	public Map<String, Map<String, ConfidenceInterval>> getAllClassIntervals(
			String class_name) {

		Map<String, Map<String, ConfidenceInterval>> output = new HashMap<String, Map<String, ConfidenceInterval>>();

		for (Map.Entry<String, Map<String, Map<String, ConfidenceInterval>>> dataset1_entry : this.confidenceIntervals
				.entrySet()) {
			String dataset1 = dataset1_entry.getKey();
			output.put(dataset1, new HashMap<String, ConfidenceInterval>());
			for (Map.Entry<String, Map<String, ConfidenceInterval>> dataset2_entry : dataset1_entry
					.getValue().entrySet()) {
				String dataset2 = dataset2_entry.getKey();
				output.get(dataset1).put(dataset2,
						dataset2_entry.getValue().get(class_name));
			}
		}

		return output;

	}

	/**
	 * Get method of confidence intervals. Only the intervals of the provided
	 * classes are returned
	 * 
	 * @param classes
	 *            Specific map of classes: Map<Class, Summarized CI?>. All CIs
	 *            of the class names are returned. If <tt>true</tt> is set of a
	 *            class the summarized interval is used. Otherwise the assigned
	 *            value CI.
	 * @return A map of contrasts of intervals for the given classes are
	 *         returned: {@code Map<population1, Map<population2, CI>>}
	 */
	public Map<String, Map<String, Map<String, ConfidenceInterval>>> getContrastsOfClasses(
			ArrayList<String> classes) {

		Map<String, Map<String, Map<String, ConfidenceInterval>>> output = new HashMap<String, Map<String, Map<String, ConfidenceInterval>>>();

		for (Map.Entry<String, Map<String, Map<String, ConfidenceInterval>>> dataset1_entry : this.confidenceIntervals
				.entrySet()) {
			String dataset1 = dataset1_entry.getKey();
			output.put(dataset1,
					new HashMap<String, Map<String, ConfidenceInterval>>());
			for (Map.Entry<String, Map<String, ConfidenceInterval>> dataset2_entry : dataset1_entry
					.getValue().entrySet()) {
				String dataset2 = dataset2_entry.getKey();
				output.get(dataset1).put(dataset2,
						new HashMap<String, ConfidenceInterval>());
				for (String class_name : classes) {
					output.get(dataset1)
							.get(dataset2)
							.put(class_name,
									dataset2_entry.getValue().get(class_name));
				}
			}
		}

		return output;
	}

	/**
	 * 
	 * This method counts for each class all Intervals that are different from
	 * Zero. The method can be used to generate a ranking of interesting classes
	 * of all populations.
	 * 
	 * @return A Map for every class where nonzero contrasts are counted under
	 *         the given confidence level. (only populationX-populationY not
	 *         populationY-populationX is counted)
	 */
	public Map<String, Double> countExcludeZero() {

		Map<String, Double> output = new HashMap<String, Double>();

		for (Map<String, ArrayList<String>> dataset1 : this.contrastsExcludeZero
				.values()) {
			for (ArrayList<String> classes : dataset1.values()) {
				for (String class_name : classes) {
					if (!output.containsKey(class_name)) {
						output.put(class_name, 0.0);
					}
					output.put(class_name, output.get(class_name) + 0.5);
				}
			}
		}
		return output;
	}

	/**
	 * 
	 * This method counts for each class all Intervals that are different from
	 * Zero. After that the counts are divided by the maximum number of
	 * contrasts for one class (n over 2).
	 * 
	 * @return A Map for every class where nonzero contrasts are counted under
	 *         the given confidence level and divided by the maximum counts of
	 *         contrasts. (only populationX-populationY not
	 *         populationY-populationX is counted)
	 */
	public Map<String, Double> getContrastScore() {

		Map<String, Double> output = countExcludeZero();

		if (this.method.getComparisons() == 0) {
			this.method.setConstants(this.numberOfPopulations,
					this.numberOfClasses);
		}

		int maxNumberOfContrasts = this.method.getComparisons();

		for (Map.Entry<String, Double> entry : output.entrySet()) {
			entry.setValue(entry.getValue() / maxNumberOfContrasts);
		}
		return output;
	}

	/**
	 * 
	 * This method counts for each class all Intervals that are different from
	 * Zero. The method can be used to generate a ranking of interesting classes
	 * of all populations.
	 * 
	 * @param classes
	 *            Specific map of classes: {@code ArrayList<String Class>}. All
	 *            given class names are returned. If <tt>true</tt> is set of a
	 *            class the summarized interval is used. Otherwise the assigned
	 *            value CI.
	 * @return A Map for every class where nonzero contrasts are counted under
	 *         the given confidence level. (only populationX-populationY not
	 *         populationY-populationX is counted)
	 */
	public Map<String, Integer> countExcludeZero(ArrayList<String> classes) {

		Map<String, Map<String, Map<String, ConfidenceInterval>>> intervals = getContrastsOfClasses(classes);

		Map<String, Integer> output = new HashMap<String, Integer>();

		for (Map<String, Map<String, ConfidenceInterval>> dataset1 : intervals
				.values()) {
			for (Map<String, ConfidenceInterval> dataset2 : dataset1.values()) {
				for (Map.Entry<String, ConfidenceInterval> class_entry : dataset2
						.entrySet()) {
					String class_name = class_entry.getKey();
					if (!output.containsKey(class_name)) {
						output.put(class_name, 0);
					}
					if (!class_entry.getValue().includeZero()) {
						output.put(class_name, output.get(class_name) + 1);
					}
				}
			}
		}

		for (Entry<String, Integer> output_entry : output.entrySet()) {
			String class_name = output_entry.getKey();
			Integer value = output_entry.getValue();
			output.put(class_name, value / 2);
		}

		return output;

	}

	/**
	 * This method computes the all p-values of 0 contrasts and reports the
	 * average p-value for every class.
	 * 
	 * @return Average p_value of 0 contrasts
	 */
	public Map<String, Double> getPValueScores() {
		Map<String, Double> scores = new HashMap<String, Double>();

		int sumOfContrasts = 0;

		for (Map<String, Map<String, ConfidenceInterval>> dataset1_value : this.confidenceIntervals
				.values()) {
			for (Map<String, ConfidenceInterval> dataset2_value : dataset1_value
					.values()) {
				sumOfContrasts++;
				for (Entry<String, ConfidenceInterval> class_entry : dataset2_value
						.entrySet()) {
					String class_name = class_entry.getKey();
					ConfidenceInterval ci = class_entry.getValue();
					if (!scores.containsKey(class_name)) {
						scores.put(class_name, ci.pValue(0.0));
					} else {
						double oldValue = scores.get(class_name);
						scores.put(class_name, oldValue + ci.pValue(0.0));
					}
				}
			}
		}
		for (Entry<String, Double> score_entry : scores.entrySet()) {
			Double new_value = score_entry.getValue() / sumOfContrasts;
			scores.put(score_entry.getKey(), new_value);
		}

		return scores;
	}

	/**
	 * This method computes and returns a score for every given class. Two sets
	 * of populations are required. The score denotes an impact factor of the
	 * class creating such split (Set1-Set2). The score ranges from 1 (high
	 * impact factor) over 0 (no impact factor/balanced factor) to -1 (negative
	 * impact/spilt is unlikely for that class).
	 * 
	 * @see Split
	 * 
	 * @param set1
	 *            String array of populations (Set 1)
	 * @param set2
	 *            String array of populations (Set 2)
	 * @param classes
	 *            Specific map of classes: {@code Map<Class, Summarized CI?>}.
	 *            All given class names are returned. If <tt>true</tt> is set of
	 *            a class the summarized interval is used. Otherwise the
	 *            assigned value CI.
	 * @return A split score for every given class. A score of a class is only
	 *         computed if the class has an interval that exclude zero in one of
	 *         the comparisons for the populations in the two sets.
	 */
	public Map<String, Double> getSplitScores(String[] set1, String[] set2,
			ArrayList<String> classes) {
		Map<String, Map<String, ArrayList<String>>> classesExcludeZero = new HashMap<String, Map<String, ArrayList<String>>>();
		// iterate over sumClassesExcludeZero and extract classes if needed
		for (Map.Entry<String, Map<String, ArrayList<String>>> population1_entry : this.contrastsExcludeZero
				.entrySet()) {
			for (Map.Entry<String, ArrayList<String>> population2_entry : population1_entry
					.getValue().entrySet()) {
				for (String class_name : population2_entry.getValue()) {
					if (classes.contains(class_name)) {
						if (!classesExcludeZero.containsKey(population1_entry
								.getKey())) {
							classesExcludeZero.put(population1_entry.getKey(),
									new HashMap<String, ArrayList<String>>());
						}
						if (!classesExcludeZero.get(population1_entry.getKey())
								.containsKey(population2_entry.getKey())) {
							classesExcludeZero.get(population1_entry.getKey())
									.put(population2_entry.getKey(),
											new ArrayList<String>());
						}
						classesExcludeZero.get(population1_entry.getKey())
								.get(population2_entry.getKey())
								.add(class_name);
					}
				}
			}
		}
		Split split = new Split(set1, set2);
		split.apply(classesExcludeZero);

		return split.getSplitScores();

	}

	/**
	 * This method computes and returns a score for every class that has a
	 * nonzero assigned value confidence interval. Two sets of populations are
	 * required. The score denotes an impact factor of the class creating such
	 * split (Set1-Set2). The score ranges from 1 (high impact factor) over 0
	 * (no impact factor/balanced factor) to -1 (negative impact/spilt is
	 * unlikely for that class).
	 * 
	 * @see Split
	 * 
	 * @param set1
	 *            String array of populations (Set 1)
	 * @param set2
	 *            String array of populations (Set 2)
	 * @return A split score for every class that has at least one assigned
	 *         value contrast that is significantly different from zero (for a
	 *         contrast if the given population of the two sets).
	 */
	public Map<String, Double> getSplitScores(String[] set1, String[] set2) {
		Split split = new Split(set1, set2);
		split.apply(this.contrastsExcludeZero);
		return split.getSplitScores();
	}

	/**
	 * This method computes and returns a score for every class that has a
	 * nonzero assigned value confidence interval. Two sets of populations are
	 * required. The score denotes an impact factor of the class creating such
	 * split (Set1-Set2). Therefore the p-value of the 0.0 value are used (equal
	 * contrast) and the average p-values of within and between sets are
	 * substracted at the end.
	 * 
	 * @param set1
	 * @param set2
	 * @return Average pValue between minus within the sets.
	 */
	public Map<String, Double> getPValueSplitScores(String[] set1, String[] set2) {
		Split split = new Split(set1, set2);
		split.apply(this.contrastsExcludeZero,
				getContrastsOfClasses(this.classesExcludeZero));
		return split.getPValueSplitScores();
	}

	/**
	 * This method computes the sopulationSizes by counting all assignments of
	 * classes per population (value parameter).
	 * 
	 * It stores the result in the populationSizes Map.
	 */
	private void computePopulationSizes() {
		this.populationSizes = new HashMap<String, Double>();
		for (Map.Entry<String, Map<String, Double>> population_entry : this.values
				.entrySet()) {
			String population = population_entry.getKey();
			double n = 0.0;
			for (Map.Entry<String, Double> class_entry : population_entry
					.getValue().entrySet()) {
				n += class_entry.getValue();
			}
			this.populationSizes.put(population, n);
		}
	}

	/**
	 * This method computes the frequencies of a given class. Therefore it
	 * divides the assigned/summarized values by the population size. It stores
	 * the two values in the double array of values map on position 2 (assigned)
	 * and 3 (summarized).
	 */
	private void computeFrequencies() {
		for (Map.Entry<String, Map<String, Double>> population_entry : this.values
				.entrySet()) {
			String population = population_entry.getKey();
			double populationSize = this.populationSizes.get(population);
			for (Map.Entry<String, Double> class_entry : population_entry
					.getValue().entrySet()) {
				class_entry.setValue(class_entry.getValue() / populationSize);
			}
		}
	}

	/**
	 * This method computes the Confidence Intervals of contrast among
	 * multinomial populations (values map) using the method of Goodman 1964
	 * (Contrast among multinomial populations).
	 * 
	 * it uses a special modification:
	 * <ul>
	 * <li>L = Math.sqrt(qchisq(alpha, degreeOfFreedom))</li>
	 * <li>Z = qnorm(alpha / (2 * G)), G = Number of used contrast</li>
	 * </ul>
	 * 
	 * If Z <= L is true then it do not use L (chisquare) for computing the
	 * quartile of alpha. Otherwise Z is used.
	 * 
	 * This modification is done, because on large degrees of freedom the
	 * chisquare distribution is quite flat. This means that the the quartile of
	 * alpha is a high number. In fact the CI gets quite large. The modification
	 * can be done, because we only use a small set of all possibly contrast
	 * (exactly 2 * (((I * (I - 1)) / 2) * J)). In fact we can reduce our
	 * complexity.
	 * 
	 */
	private void computeConfidenceIntervals() {

		this.confidenceIntervals = new HashMap<String, Map<String, Map<String, ConfidenceInterval>>>();
		this.contrastsIncludeZero = new HashMap<String, Map<String, ArrayList<String>>>();
		this.contrastsExcludeZero = new HashMap<String, Map<String, ArrayList<String>>>();
		this.classesIncludeZero = new ArrayList<String>();
		this.classesExcludeZero = new ArrayList<String>();

		double percentile = this.method.getPercentile(this.numberOfPopulations,
				this.numberOfClasses);

		for (Map.Entry<String, Map<String, Double>> population_1_entry : this.values
				.entrySet()) {
			String populationName1 = population_1_entry.getKey();
			// set map
			this.confidenceIntervals.put(populationName1,
					new HashMap<String, Map<String, ConfidenceInterval>>());
			this.contrastsIncludeZero.put(populationName1,
					new HashMap<String, ArrayList<String>>());
			this.contrastsExcludeZero.put(populationName1,
					new HashMap<String, ArrayList<String>>());

			for (Map.Entry<String, Map<String, Double>> population_2_entry : this.values
					.entrySet()) {
				String populationName2 = population_2_entry.getKey();
				if (populationName1.equals(populationName2)) {
					continue;
				}
				// set map
				this.confidenceIntervals.get(populationName1).put(
						populationName2,
						new HashMap<String, ConfidenceInterval>());
				this.contrastsIncludeZero.get(populationName1).put(
						populationName2, new ArrayList<String>());
				this.contrastsExcludeZero.get(populationName1).put(
						populationName2, new ArrayList<String>());
				// compute contrast & delta
				for (Map.Entry<String, Double> class_entry : population_1_entry
						.getValue().entrySet()) {
					String class_name = class_entry.getKey();
					double p1 = class_entry.getValue();
					double p2 = 0.0;
					if (population_2_entry.getValue().containsKey(class_name)) {
						p2 = population_2_entry.getValue().get(class_name);
					}
					double variance = (p1 - (p1 * p1))
							/ this.populationSizes.get(populationName1)
							+ (p2 - (p2 * p2))
							/ this.populationSizes.get(populationName2);

					double standardDeviation = Math.sqrt(variance);

					ConfidenceInterval ci = new ConfidenceInterval(p1 - p2,
							standardDeviation * percentile, standardDeviation,
							this.method);

					// set include/exclude zero lists
					if (ci.includeZero()) {
						this.contrastsIncludeZero.get(populationName1)
								.get(populationName2).add(class_name);
						this.classesIncludeZero.add(class_name);
					} else {
						this.contrastsExcludeZero.get(populationName1)
								.get(populationName2).add(class_name);
						this.classesExcludeZero.add(class_name);
					}

					this.confidenceIntervals.get(populationName1)
							.get(populationName2).put(class_name, ci);

				}
			}

		}

		if (this.method.getMethod().equals("HOLM-BONFERRONI")) {

			this.method.holmBonferroniCorrection(this.confidenceIntervals);
			for (Map.Entry<String, Map<String, ArrayList<String>>> dataset1_entry : this.contrastsExcludeZero
					.entrySet()) {
				String dataset1 = dataset1_entry.getKey();
				for (Map.Entry<String, ArrayList<String>> dataset2_entry : dataset1_entry
						.getValue().entrySet()) {
					String dataset2 = dataset2_entry.getKey();
					ArrayList<String> remove_classes = new ArrayList<String>();
					for (String class_name : dataset2_entry.getValue()) {
						if (this.getClassInterval(dataset1, dataset2,
								class_name).pValue(0.0) >= this.method
								.getPValueBorder()) {
							remove_classes.add(class_name);
						}
					}
					for (String class_name : remove_classes) {
						dataset2_entry.getValue().remove(class_name);
						if (!this.contrastsIncludeZero.get(dataset1)
								.get(dataset2).contains(class_name)) {
							this.contrastsIncludeZero.get(dataset1)
									.get(dataset2).add(class_name);
						}
					}
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {

		return this.confidenceIntervals.toString();
	}

}